import os

from pyfileconf.pipelines.models.config import FunctionConfig
from pyfileconf.logic.get import _get_public_name_or_special_name
from pyfileconf.pipelines.models.interfaces import ObjectViewOrCollection
from pyfileconf.views.object import ObjectView

from pyfileconf.basemodels.collection import Collection

class PipelineCollection(Collection):

    def _set_name_map(self):
        pipeline_map = {}
        for pipeline_or_collection in self:
            pipeline_name = _get_public_name_or_special_name(pipeline_or_collection)
            pipeline_map[pipeline_name] = pipeline_or_collection
        self.name_dict = pipeline_map

    def _transform_item(self, item):
        """
        Is called on each item when adding items to collection. Should handle whether the item
        is an actual item or another collection. Must return the item or collection.

        If not overridden, will just store items as is.

        Returns: item or Collection

        """
        if isinstance(item, PipelineCollection):
            # no processing needed for collections, just items
            return item
        # If item, convert to object view
        return ObjectView.from_ast_and_imports(item, self.imports)

    def _output_config_files(self):

        if not os.path.exists(self.basepath):
            os.makedirs(self.basepath)

        self._output_section_config_file()
        [self._output_config_file(item) for item in self]

    def _output_config_file(self, item: ObjectViewOrCollection):
        if isinstance(item, PipelineCollection):
            # if collection, recursively call creating config files
            return item._output_config_files()

        if not isinstance(item, ObjectView):
            raise ValueError(f'did not initialize pipeline dict correctly. expected ObjectView, got {item} of '
                             f'type {type(item)}')

        # Dealing with ObjectView
        item_filepath = os.path.join(self.basepath, item.output_name + '.py')

        if os.path.exists(item_filepath):
            # if config file already exists, load confguration from file, use to update function defaults
            existing_config = FunctionConfig.from_file(item_filepath, item.output_name)
        else:
            existing_config = FunctionConfig()

        item_config = item.default_config.copy()
        item_config.update(existing_config) # override function defaults with any settings from file
        item_config.to_file(item_filepath)

    def _output_section_config_file(self):
        """
        creates a blank config file for the section
        """
        outpath = os.path.join(self.basepath, 'section.py')

        if os.path.exists(outpath):
            # Never overwrite section config.
            return

        with open(outpath, 'w') as f:
            f.write('\n')






from pyfileconf.io.file.load.lazy.pipelineast import PipelineAstLoader
from pyfileconf.io.file.load.parsers.collections import extract_collection_from_ast

class PipelineDictLoader(PipelineAstLoader):

    def load(self):
        # Get pipeline_dict_ast
        super().load()

        # Convert ast dicts and ast lists to normal dicts and lists, store as self.pipeline_dict
        self._pipeline_dict_assign_to_dict()

        return self.pipeline_dict

    @property
    def pipeline_dict(self):
        return self._try_getattr_else_load('_pipeline_dict')

    def _pipeline_dict_assign_to_dict(self):
        """
        Iterates through ast tree of pipeline dict, producing a dict/list structure, while leaving
        other objects as ast representations

        Returns:

        """
        if self.pipeline_dict_assign is not None:
            self._pipeline_dict = extract_collection_from_ast(self.pipeline_dict_assign.to_ast())
        else:
            self._pipeline_dict = {}



import os
from copy import deepcopy

from pyfileconf import Selector, PipelineManager
from pyfileconf.sectionpath.sectionpath import SectionPath
from pyfileconf.context import context
from pyfileconf.selector.models.itemview import ItemView
from tests.input_files.amodule import SecondExampleClass
from tests.input_files.mypackage.cmodule import ExampleClass, ExampleClassProtocol
from tests.test_pipeline_manager.base import PipelineManagerTestBase, CLASS_CONFIG_DICT_LIST, SAME_CLASS_CONFIG_DICT_LIST, \
    DIFFERENT_CLASS_CONFIG_DICT_LIST


class TestItemView(PipelineManagerTestBase):

    def assert_valid_function_iv(self, iv: ItemView, pipeline_manager: PipelineManager):
        assert isinstance(iv, ItemView)
        assert iv() == pipeline_manager.run(iv)
        assert iv.item() == pipeline_manager.get(iv)()
        iv2 = ItemView.from_section_path_str(iv._section_path_str)
        assert iv == iv2
        assert iv2 in [iv]
        assert iv == list({iv})[0]
        assert {iv: 5}[iv] == 5
        assert hash(iv) == hash(iv.section_path_str)

    def assert_valid_class_iv(self, iv: ItemView, pipeline_manager: PipelineManager):
        assert isinstance(iv, ItemView)
        assert iv() == pipeline_manager.get(iv)()
        assert iv()() == pipeline_manager.run(iv)
        assert iv.item() == pipeline_manager.get(iv)()
        iv2 = ItemView.from_section_path_str(iv._section_path_str)
        assert iv == iv2
        assert iv2 in [iv]
        assert iv == list({iv})[0]
        assert {iv: 5}[iv] == 5
        assert hash(iv) == hash(iv.section_path_str)

    def assert_valid_specific_class_iv(self, iv: ItemView, pipeline_manager: PipelineManager):
        assert isinstance(iv, ItemView)
        assert iv() == pipeline_manager.get(iv)()
        assert iv() == pipeline_manager.run(iv)
        assert iv.a == pipeline_manager.get(iv).a
        assert iv.item() == pipeline_manager.get(iv)()
        iv2 = ItemView.from_section_path_str(iv._section_path_str)
        assert iv == iv2
        assert iv2 in [iv]
        assert iv == list({iv})[0]
        assert {iv: 5}[iv] == 5
        assert hash(iv) == hash(iv.section_path_str)
        assert isinstance(iv, ExampleClassProtocol)

    def test_function_iv_from_selector(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.a_function
        self.assert_valid_function_iv(iv, pipeline_manager)

    def test_function_iv_from_str(self):
        self.write_a_function_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        iv = ItemView.from_section_path_str('test_pipeline_manager.stuff.a_function')
        self.assert_valid_function_iv(iv, pipeline_manager)

    def test_class_iv_from_selector(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.stuff.ExampleClass
        self.assert_valid_class_iv(iv, pipeline_manager)

    def test_class_iv_from_str(self):
        self.write_example_class_to_pipeline_dict_file()
        pipeline_manager = self.create_pm()
        pipeline_manager.load()
        iv = ItemView.from_section_path_str('test_pipeline_manager.stuff.ExampleClass')
        self.assert_valid_class_iv(iv, pipeline_manager)

    def test_specific_class_iv_from_selector(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        sel = Selector()
        iv = sel.test_pipeline_manager.example_class.stuff.data
        self.assert_valid_specific_class_iv(iv, pipeline_manager)

    def test_specific_class_iv_from_str(self):
        self.write_example_class_dict_to_file()
        pipeline_manager = self.create_pm(
            specific_class_config_dicts=CLASS_CONFIG_DICT_LIST
        )
        pipeline_manager.load()
        iv = ItemView.from_section_path_str('test_pipeline_manager.example_class.stuff.data')
        self.assert_valid_specific_class_iv(iv, pipeline_manager)





import os
from typing import TYPE_CHECKING, List, Dict, Type, Optional, Sequence

from pyfileconf.assignments.models.statement import AssignmentStatement
from pyfileconf.imports.models.statements.interfaces import AnyImportStatement

if TYPE_CHECKING:
    from pyfileconf.basemodels.config import ConfigBase

from pyfileconf.sectionpath.sectionpath import _strip_py
from pyfileconf.assignments.models.container import AssignmentStatementContainer
from pyfileconf.io.file.interfaces.config import ConfigFileInterface


class ConfigFileBase:

    ##### Scaffolding functions or attributes. Need to override when subclassing  ####

    # lines to always import. pass import objects
    always_imports: List[AnyImportStatement] = []

    # assignment lines to always include at beginning. pass assignment objects
    always_assigns: List[AssignmentStatement] = []

    # always assign dict, where assigns will get added if item name matches dict key
    always_assign_with_names_dict: Dict[str, List[AssignmentStatement]] = {}

    # class to use for interfacing with file
    interface_class = ConfigFileInterface

    ##### Base class functions and attributes below. Shouldn't usually need to override in subclassing #####

    def __init__(self, filepath: str, name: str=None, klass: Optional[Type] = None,
                 always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None):
        self.interface = self.interface_class(filepath)

        # TODO [#23]: check if setting filepath in ConfigFileBase.__init__ had side effects
        #
        # added this because filepath was being set after object creation in
        # `pyfileconf.basemodels.config.ConfigBase.to_file` and was causing mypy errors. Check
        # to ensure this didn't cause any issues.
        self.filepath = filepath

        if name is None:
            name = _strip_py(os.path.basename(filepath))

        self.name = name
        self.klass = klass
        self.always_import_strs = always_import_strs
        self.always_assign_strs = always_assign_strs

    def load(self, config_class: type = None) -> 'ConfigBase':
        from pyfileconf.basemodels.config import ConfigBase
        config_dict, annotation_dict = self.interface.load()

        if config_class is None:
            config_class = ConfigBase

        return config_class(
            d=config_dict,
            annotations=annotation_dict,
            imports=self.interface.imports,
            _file=self,
            name=self.name,
            klass=self.klass,
            always_import_strs=self.always_import_strs,
            always_assign_strs=self.always_assign_strs,
        )

    def save(self, config: 'ConfigBase') -> None:
        self._add_always_imports_and_assigns_to_config(config)
        self.interface.save(config)

    def _add_always_imports_and_assigns_to_config(self, config: 'ConfigBase'):
        """
        Note: inplace
        """
        # Add always imports
        [config.imports.add_if_missing(imp) for imp in self.always_imports]

        # # Check if there are any extra assigns for items with this name
        always_assigns = self.always_assigns.copy()
        if self.name in self.always_assign_with_names_dict:
            always_assigns.extend(self.always_assign_with_names_dict[self.name])

        # Add always assigns
        # First handle begin assigns
        begin_assigns = AssignmentStatementContainer(
            [assign for assign in always_assigns if assign.prefer_beginning]
        )
        config.begin_assignments = begin_assigns
        # Now handle the rest
        # First get always assigns, annotations as dict
        other_always_assigns = AssignmentStatementContainer(
            [assign for assign in always_assigns if not assign.prefer_beginning]
        )
        always_defaults, always_annotations = other_always_assigns.to_default_dict_and_annotation_dict()
        # Select assigns, annotations which are not already defined in config
        new_defaults = {key: value for key, value in always_defaults.items() if key not in config}
        new_annotations = {key: value for key, value in always_annotations.items() if key not in config.annotations}
        # Add to config
        config.update(new_defaults)
        config.annotations.update(new_annotations)



from typing import Optional, Type, Any, Sequence, Union

from pyfileconf.basemodels.config import ConfigBase
from pyfileconf.data.models.file import SpecificClassConfigFile, ConfigFileBase
from pyfileconf.imports.logic.load.klass import class_function_args_as_dict
from pyfileconf.imports.models.statements.container import ImportStatementContainer
from pyfileconf.assignments.models.container import AssignmentStatementContainer
from pyfileconf.data.models.astitems import ast_str, ast_none


class SpecificClassConfig(ConfigBase):
    config_file_class: Union[SpecificClassConfigFile, Type[ConfigFileBase]]  # type: ignore

    def __init__(self, d: dict = None, name: str = None, annotations: dict = None,
                 imports: ImportStatementContainer = None,
                 _file: ConfigFileBase = None, begin_assignments: AssignmentStatementContainer = None,
                 always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None, klass: Optional[Type] = None,
                 file_path: Optional[str] = None, **kwargs):

        self.config_file_class = SpecificClassConfigFile(
            file_path,  # type: ignore
            name=name,
            klass=klass,
            always_import_strs=always_import_strs,
            always_assign_strs=always_assign_strs
        )

        super().__init__(
            d=d,
            name=name,
            annotations=annotations,
            imports=imports,
            _file=_file,
            begin_assignments=begin_assignments,
            klass=klass,
            always_import_strs=always_import_strs,
            always_assign_strs=always_assign_strs,
            **kwargs
        )



    @classmethod
    def from_obj(cls, obj: Any, klass: Type, name: str=None, imports: ImportStatementContainer = None,
                 always_import_strs: Optional[Sequence[str]] = None,
                 always_assign_strs: Optional[Sequence[str]] = None, file_path: Optional[str] = None,
                 key_attr: str = 'name'):
        # Initialize a blank config dictionary
        config_dict = class_function_args_as_dict(klass)

        # Replace obj None with ast None
        config_dict = {key: value if value is not None else ast_none for key, value in config_dict.items()}

        # Fill blank config dict
        for config_attr in config_dict:
            if not _obj_attr_is_none(obj, config_attr):
                config_dict[config_attr] = getattr(obj, config_attr)

        # Special handling for name, which will be set even before creating file
        if isinstance(config_dict[key_attr], str):
            config_dict[key_attr] = ast_str(config_dict[key_attr])  # convert str to ast

        return cls(
            config_dict,
            name=name,
            imports=imports,
            klass=klass,
            always_import_strs=always_import_strs,
            always_assign_strs=always_assign_strs,
            file_path=file_path
        )




def _obj_attr_is_none(obj: Any, source_attr: str) -> bool:
    source_value = getattr(obj, source_attr, None)
    if source_value is None:
        return True

    return False



from typing import Type, List

from typing_extensions import TypedDict


SpecificClassConfigDict = TypedDict(
    "SpecificClassConfigDict",
    {
        "name": str,
        "class": Type,
        "always_import_strs": List[str],
        "always_assign_strs": List[str],
        "key_attr": str,
        "execute_attr": str,
    },
    total=False
)



from typing import Union

from pyfileconf.config.models.file import ActiveFunctionConfigFile
from pyfileconf.config.models.section import ConfigSection, ActiveFunctionConfig


ConfigSectionOrConfig = Union[ConfigSection, ActiveFunctionConfigFile, ActiveFunctionConfig]


from contextlib import contextmanager
import datetime
import sys
import os

@contextmanager
def stdout_also_logged(log_folder: str = None):
    stdout_logger = Logger(log_folder)
    sys.stdout = stdout_logger  # type: ignore
    yield log_folder
    sys.stdout = stdout_logger.terminal
    stdout_logger.log.close()


class Logger(object):
    def __init__(self, log_folder: str = None):
        self.terminal = sys.stdout
        log_path = _log_filename() if log_folder is None else os.path.join(log_folder, _log_filename())
        self.log = open(log_path, "a", encoding='utf8')

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        #this flush method is needed for python 3 compatibility.
        #this handles the flush command by doing nothing.
        #you might want to specify some extra behavior here.
        pass

def _log_filename():
    creation_time = str(datetime.datetime.now().replace(microsecond=0)).replace(':', '.')
    return f'{creation_time}.log'


from pyfileconf.io.file.load.lazy.base.impassign import ImportAssignmentLazyLoader
from pyfileconf.io.file.load.parsers.assign import extract_assignment_from_ast_by_name

class SpecificClassDictAstLoader(ImportAssignmentLazyLoader):

    def load(self):
        # Get ast, imports, assigns
        super().register()

        # Store pipeline dict assignment
        self._class_dict_assign = extract_assignment_from_ast_by_name(self._ast, 'class_dict')

    @property
    def class_dict_assign(self):
        return self._try_getattr_else_load('_class_dict_assign')

    def _try_getattr_else_load(self, attr: str):
        return self._try_getattr_else_call_func(attr, self.load)


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.basemodels.config import ConfigBase

from pyfileconf.io.file.load.lazy.config import ConfigFileLoader
from pyfileconf.io.file.write.config import ConfigFileStr

class ConfigFileInterface(ConfigFileLoader):

    def save(self, config: 'ConfigBase'):
        file_str_obj = ConfigFileStr(
            config,
            existing_assigns=self.assigns,
            existing_imports=self.imports,
            existing_body=self.assignment_body
        )

        with open(self.filepath, 'w', newline='\n', encoding='utf8') as f:
            f.write(file_str_obj.file_str)


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyfileconf.basemodels.config import ConfigBase

from pyfileconf.assignments.models.container import AssignmentStatementContainer
from pyfileconf.io.file.load.lazy.datadict import SpecificClassDictLoader
from pyfileconf.io.file.write.base import FileStr


class SpecificClassDictInterface(SpecificClassDictLoader):

    def save(self, config: 'ConfigBase'):
        # Set existing assigns and body to empty because should be completely replaced by new specific class dict
        file_str_obj = FileStr(
            config,
            existing_assigns=AssignmentStatementContainer([]),
            existing_imports=self.imports,
            existing_body=[]
        )

        with open(self.filepath, 'w', newline='\n', encoding='utf8') as f:
            f.write(file_str_obj.file_str)